@@ -5,16 +5,12 @@ from loguru import logger
 from typing import Callable, Dict
 
 default_config = {
-    'env': 0, # {'dev': 0, 'production': 1}
     'tcp-port': 4004,
     'tcp-ip': 'ngrok',
     'tcp-timeout': 10,
-    'log-level': 'INFO',
+    'max-workers': 32,
     'echo-responses': False,
     'media-quality': 'high',
-    'ngrok-path': None,
-    'ngrok-token': None,
-    'ngrok-config-path': None,
     'telegram-token': None,
 }
 
@@ -31,17 +27,15 @@ def load_config(config_file : str, config_path : str):
         print(f"Please edit [{config_file}]")
         sys.exit(1)
     config = {'config-file': config_file}
-    try:
-        with open(config_file, 'r') as conf:
-            config |= json.load(conf)
-    except json.decoder.JSONDecodeError:
-        return None
+    with open(config_file, 'r') as conf:
+        config |= json.load(conf)
     return config
 
 @logger.catch
 def save_config(config :Dict):
     """config-file must be in the dictionary"""
-    if conffile := config.get('config-file'):
+    if config.get('config-file'):
+        conffile = config.pop('config-file')
         with open(conffile, 'w') as conf:
             json.dump(config, conf, indent=3)
             conf.flush()
