@@ -1,16 +1,14 @@
+import re
 import sys
 import json
 import time
 import loguru
 import asyncio
-import aiohttp
 import requests
 import threading
-from typing import Callable
-from typing import Dict, Any
 from queue import Queue, Empty
 from contextlib import contextmanager
-from pyngrok import ngrok, conf as ngrokconf
+from typing import Dict, Any, Callable, Tuple
 from concurrent.futures import ThreadPoolExecutor
 #
 from autogram import *
@@ -22,17 +20,20 @@ class Autogram:
     api_url = 'https://api.telegram.org/'
 
     def __init__(self, config: Dict):
+        self.link_patt = re.compile('^(https?):\\/\\/[^\\s/$.?#].[^\\s]*$')
         self.token :str|None = config.get('telegram-token')
         if not self.token:
             print('Missing bot token!')
             sys.exit(1)
         #
+        self.ip = None
         self.config = config
         self._initialize_()
 
     def _initialize_(self):
         self.webhook = None
         self.host = '0.0.0.0'
+        self.edited_ = set()
         self.deleted_ = set()
         self.public_ip = None
         self.update_offset = 0
@@ -40,12 +41,14 @@ class Autogram:
         self.ngrok_tunnel = None
         self.httpRoutines = list()
         self.httpRequests = Queue()
+        self.worker_threads = list()
         self.failing_endpoints = list()
+        self.session = requests.session()
         self.terminate = threading.Event()
         self.port = self.config['tcp-port']
-        self.executor = ThreadPoolExecutor()
+        self.executor = ThreadPoolExecutor(max_workers=self.config['max-workers'])
+        self.timeout = self.config['tcp-timeout'] or 10
         self.base_url = f"{Autogram.api_url}bot{self.token}"
-        self.timeout = aiohttp.ClientTimeout(self.config['tcp-timeout'])
         #
         self.guard = {
             'lock': threading.Lock(),
@@ -53,28 +56,10 @@ class Autogram:
             'thread': None
         }
         #
-        self.worker_threads = {
-            'normal': list(),
-            'high': list(),
-        }
         self.locks = {
             'session': threading.Lock(),
         }
-        # 
-        logger_format = (
-            "<green>{time:DD/MM/YYYY HH:mm:ss}</green> | "
-            "<level>{level: <8}</level>|"
-            "<cyan>{line}</cyan>:<cyan>{name}</cyan>:<cyan>{function}</cyan> | "
-            "<level>{message}</level>"
-        )
-        loguru.logger.remove()
         #
-        lvl = 'DEBUG'
-        if env := self.config.get('env'):
-            if env == 1:
-                lvl = self.config.get('log-level') or 'DEBUG'
-        #
-        loguru.logger.add(sys.stderr, format=logger_format, level=lvl)
         self.logger = loguru.logger
         return
 
@@ -85,49 +70,38 @@ class Autogram:
             return 1
         return 0
 
-    @loguru.logger.catch
     def send_online(self) -> threading.Thread:
         """Get this bot online in a separate daemon thread."""
-        if not self.token:
-            raise RuntimeError("Send online without token?!")
-        if public_ip := self.config['tcp-ip']:
+        try:
+            if not self.token:
+                raise RuntimeError("No telegram token provided!")
+            self.getMe()
+        except Exception as e:
+            raise RuntimeError(str(e))
+        #
+        if (public_ip := self.config['tcp-ip']) and not self.terminate.is_set():
+            if not re.search(self.link_patt, public_ip):
+                raise RuntimeError('Unknown public url format!')
+            self.public_ip = public_ip
             hookPath = self.token.split(":")[-1].lower()
             @post(f'/{hookPath}')
             def hookHandler():
                 self.updateRouter(request.json)
                 response.content_type = 'application/json'
                 return json.dumps({'ok': True})
-            # 
+            #
             def runServer(server: Any):
                 run(server=server, quiet=True)
-            # 
+            #
             server = MyServer(host=self.host,port=self.port)
             serv_thread = threading.Thread(target=runServer, args=(server,))
             serv_thread.name = 'Autogram::Bottle'
             serv_thread.daemon = True
             serv_thread.start()
-            # check public ip availability
-            if public_ip == 'ngrok':
-                ngrok_config = ngrokconf.PyngrokConfig(
-                    ngrok_version='v3',
-                    ngrok_path= self.config.get('ngrok-path'),
-                    auth_token= self.config.get('ngrok-token'),
-                    config_path= self.config.get('ngrok-config-path')
-                )
-                ngrokconf.set_default(ngrok_config)
-                try:
-                    self.ngrok_tunnel = ngrok.connect()
-                    ngrok.get_ngrok_process().stop_monitor_thread()
-                    public_ip = self.ngrok_tunnel.public_url
-                except Exception as e:
-                    self.logger.critical(e)
-                    self.shutdown()
             #
-            self.public_ip = public_ip
             if not self.terminate.is_set():
                 self.webhook = f"{self.public_ip}/{hookPath}"
-                self.logger.debug(f'Webhook: {self.webhook}')
-        # wrap and start main_loop
+        # wrap and start
         def launch():
             if self.terminate.is_set():
                 return
@@ -152,7 +126,6 @@ class Autogram:
     async def main_loop(self):
         """Main control loop"""
         processor = asyncio.create_task(self.aioWebRequest())
-        self.httpRequests.put((f'{self.base_url}/getMe',None,self.getMe))
         if self.webhook:
             url = f'{self.base_url}/setWebhook'
             self.httpRequests.put((url,{
@@ -167,7 +140,6 @@ class Autogram:
             self.deleteWebhook()
             self.getWebhookInfo(check_webhook)
         #
-        await asyncio.sleep(0)    # allow getMe to run
         await processor
         return
 
@@ -212,7 +184,7 @@ class Autogram:
                 parser = chatMember
             elif payload:= update.get(chatJoinRequest.name):
                 parser = chatJoinRequest
-            # 
+            #
             if not parser:
                 return
             # todo: implement all update types then allow through
@@ -223,7 +195,7 @@ class Autogram:
             #
             parser.autogram = self
             self.toThread(parser,payload)
-        # 
+        #
         if type(res) == list:
             for update in res:
                 if update['update_id'] >= self.update_offset:
@@ -235,75 +207,45 @@ class Autogram:
 
     def popWorkers(self, workers: list):
         for item in workers:
-            key, task = item
-            self.worker_threads[key].remove(task)
-            continue
+            if item in self.worker_threads:
+                self.worker_threads.remove(item)
         return
 
     @loguru.logger.catch
     def threadGuard(self):
-        workers = 0
         to_remove = list()
         self.guard['lock'].acquire()
         while not self.terminate.is_set():
             try:
-                worker = self.guard['pending'].get(timeout=3)
-                priority, work = worker
-                #
-                self.worker_threads[priority].append(work)
-                workers += 1
+                work = self.guard['pending'].get(timeout=5)
+                self.worker_threads.append(work)
             except Empty:
-                continue
+                pass
             #
-            for key in self.worker_threads:
-                for task in self.worker_threads[key]:
-                    #
-                    _ , tsk, cb, errh = task
-                    if not tsk.done():
-                        continue
-                    # fetch results or exceptions
-                    try:
-                        result = tsk.result()
-                        if cb:
-                            self.toThread(cb, result)
-                    except Exception as e:
-                        if errh:
-                            errh(e)
-                        else:
-                            self.logger.exception(e)
-                    to_remove.append((key, task))
-                    workers -= 1
+            for task in self.worker_threads:
+                nm , tsk, errh = task
+                if not tsk.done():
+                    continue
+                # fetch results or exceptions
+                if error := tsk.exception():
+                    if errh:
+                        errh(error)
+                    else:
+                        self.logger.exception(e)
+                to_remove.append(task)
             #
             self.popWorkers(to_remove)
             to_remove.clear()
         # shutdown executor
-        if not workers:
-            self.guard['lock'].release()
-            return
-        elif self.worker_threads['high'] or self.worker_threads['normal']:
-            self.logger.debug("Terminating worker threads...")
-            #
-            for task in self.worker_threads['normal']:
-                tsk_id, tsk, cb, _ = task
-                if not tsk.done() and not tsk.running():
-                    if tsk.cancel():
-                        self.logger.debug(f"[{tsk_id}] canceled.")
-                    else:
-                        self.logger.debug(f"[{tsk_id}] rejected cancel request.")
-                elif done := tsk.done() and cb and not tsk.exception():
-                    self.logger.debug(f"[{tsk_id}] completed but was ignored on termination.")
-                elif done and (e := tsk.exception()):
-                    self.logger.exception(e)
-                continue
-            #
-            for worker in self.worker_threads['high']:
-                tsk_id, tsk, cb, _ = worker
+        if self.worker_threads:
+            self.logger.debug("[ThreadGuard] terminating threads...")
+            for task in self.worker_threads:
+                tsk_id, tsk, errh = task
                 if tsk.done():
-                    to_remove.append(('high',worker))
-                    # propagate result
-                    if cb and not tsk.exception():
-                        cb(tsk.result())
-                    elif e := tsk.exception():
+                    to_remove.append(task)
+                    if e := tsk.exception() and errh:
+                        errh(e)
+                    else:
                         self.logger.exception(e)
                         self.logger.debug(f"[{tsk_id}] finished with an error.")
                 elif not tsk.running():
@@ -312,16 +254,15 @@ class Autogram:
                     else:
                         self.logger.debug(f"[{tsk_id}] : Cancel failed!")
                 else:
-                    self.logger.debug(f"[{tsk_id}] : thread is busy!")
+                    self.logger.debug(f"[{tsk_id}] : thread busy!")
                 continue
             self.executor.shutdown(wait=False, cancel_futures=True)
-        #
-        self.logger.debug("Tasks terminated.")
+        self.logger.debug("[ThreadGuard] threads terminated.")
         self.guard['lock'].release()
         return
 
     @loguru.logger.catch()
-    def toThread(self, *args, callback :Callable|None = None, errHandler :Callable|None = None, priority :str|None = None):
+    def toThread(self, *args, callback :Callable|None = None, errHandler :Callable|None = None):
         if self.locks['session'].locked():
             self.logger.debug(f"[{args[0].__name__}] : Blocked execution")
             return
@@ -332,18 +273,16 @@ class Autogram:
             self.guard['thread'].start()
         # append worker to priority group
         result = False
-        if not priority or (result := (priority not in list(self.worker_threads.keys()))):
-            if result:
-                self.logger.debug(f"Undefined task priority: {priority}: {args}")
-            priority = 'normal'
         #
         tsk_id = args[0].__name__
         try:
             task = self.executor.submit(*args)
-            self.guard['pending'].put((priority, (tsk_id, task, callback, errHandler)))
+            if callback:
+                task.add_done_callback(callback)
+            self.guard['pending'].put((tsk_id, task, errHandler))
             return task
         except RuntimeError:
-            self.terminate.set()
+            self.shutdown()
         return
 
     @contextmanager
@@ -365,7 +304,7 @@ class Autogram:
                         yield None
                     except Exception as e:
                         self.logger.exception(e)
-                        self.terminate.set()
+                        self.shutdown()
                         yield None
                 else:
                     yield self.httpRequests.get(block=False)
@@ -403,14 +342,14 @@ class Autogram:
                     elif self.config.get('echo-responses'):
                             self.logger.debug(payload)
                     continue
-                elif resp.status == 401:
+                elif resp.status_code == 401:
                     self.logger.critical("Invalid token. Closing...")
                     self.shutdown()
                 elif endpoint not in self.failing_endpoints:
                     if payload:
-                        self.logger.critical(f"[{endpoint}] HTTP{resp.status} : {endpoint} : {payload}")
+                        self.logger.critical(f"[{endpoint}] HTTP{resp.status_code} : {endpoint} : {payload}")
                     else:
-                        self.logger.critical(f"[{endpoint}] HTTP{resp.status} : {outgoing}")
+                        self.logger.critical(f"[{endpoint}] HTTP{resp.status_code} : {outgoing}")
                 # ignore repeated output
                 self.failing_endpoints.append(endpoint)
                 continue
@@ -420,7 +359,7 @@ class Autogram:
     @loguru.logger.catch()
     async def aioWebRequest(self):
         """Make asynchronous requests to the Telegram API"""
-        async with aiohttp.ClientSession(timeout=self.timeout) as session:
+        with requests.session() as session:
             self.failed = None
             asyncio.create_task(self.httpHandler())
             #
@@ -445,7 +384,7 @@ class Autogram:
                         'params': {
                             "limit": 81,
                             "offset": self.update_offset,
-                            "timeout": self.timeout.total,
+                            "timeout": self.timeout,
                         }
                     }
                     if not kw.get('params'):
@@ -458,26 +397,18 @@ class Autogram:
                     #
                     error_detected = None
                     try:
-                        async with session.get(link,**kw) as resp:
-                            data = None
-                            if resp.ok:
-                                data = await resp.json()
-                            self.logger.debug(f'[{resp.status}] GET /{link.split("/")[-1]}')
+                        with session.get(link, **kw, timeout=None) as resp:
+                            data = resp.json() or resp.content
+                            self.logger.debug(f'[{resp.status_code}] GET /{link.split("/")[-1]}')
                             self.httpRoutines.append(((resp, data), request))
                             await asyncio.sleep(0)
                     except KeyboardInterrupt:
-                        self.terminate.set()
-                    except aiohttp.ClientConnectorError as e:
-                        error_detected = e
-                    except aiohttp.ClientOSError as e:
-                        error_detected = e
-                    except asyncio.TimeoutError as e:
-                        error_detected = e
+                        self.shutdown()
                     except RuntimeError as e:
                         error_detected = e
                     except Exception as e:
                         error_detected = e
-                        self.terminate.set()
+                        self.shutdown()
                         self.logger.exception(e)
                     finally:
                         if error_detected:
@@ -486,33 +417,34 @@ class Autogram:
                 await asyncio.sleep(0)
 
     @loguru.logger.catch()
-    def webRequest(self, url: str, params={}, files=None):
+    def webRequest(self, url: str, params={}, files=None) -> Tuple[Any,Any]:
         res = None
         params = params or {}
         # send request
         try:
-            if files:
-                res = requests.get(url,params=params,files=files)
-            else:
-                res = requests.get(url,params=params)
+            with requests.session() as session:
+                if files:
+                    res = session.get(url,params=params,files=files, timeout=self.timeout)
+                else:
+                    res = session.get(url, params =params, timeout =self.timeout)
             #
             if res.ok:
                 return True, json.loads(res.text)['result']
+            self.logger.critical(f"{url.split('/')[-1]} {res.status_code}: {res.content}")
+        except requests.exceptions.ConnectionError:
+            res = 'Connection Error. Aborting.'
         except Exception as e:
-            self.logger.exception(e)
-        finally:
-            return False, res, url
+            if not res:
+                res = e
+        return False, res
 
     def shutdown(self, callback :Callable|None = None):
         """callback: your exit function that takes `msg : str`"""
         if self.locks['session'].locked() or self.terminate.is_set():
             return
         # block further updates
-        try:
-            ngrok.disconnect(self.public_ip)
-        except:
-            pass
-        #
+        if self.ip:
+            self.ip.terminate.set()
         # start termination
         self.terminate.set()
         if callback:
@@ -525,7 +457,6 @@ class Autogram:
             self.locks['session'].acquire()
         # terminate and wait for threadGuard
         self.logger.info('Autogram::terminating...')
-        self.terminate.set()
         if self.guard['lock'].locked():
             self.guard['thread'].join()
         #
@@ -533,10 +464,17 @@ class Autogram:
         return
 
     #***** start API calls *****#
-    def getMe(self, me: Dict):
-        """receive and parse getMe request."""
-        for k,v in me.items():
-            setattr(self, k, v)
+    def getMe(self):
+        """Authenticate bot"""
+        url = f'{self.base_url}/getMe'
+        ok, res = self.webRequest(url)
+        if not ok:
+            raise RuntimeError("getMe() failed!")
+            return False
+        else:
+            for k,v in res.items():
+                setattr(self, k, v)
+        return True
 
     @loguru.logger.catch()
     def downloadFile(self, file_path: str):
@@ -579,13 +517,24 @@ class Autogram:
         callback = None
         if 'callback' in kwargs.keys():
             callback = kwargs.pop('callback')
+        #
         url = f'{self.base_url}/sendMessage'
+        if kwargs.get('urgent'):
+            kwargs.pop('urgent')
+            params = {
+                'params': {
+                    'chat_id': chat_id,
+                    'text': text,
+                } | kwargs
+            }
+            return self.webRequest(url, **params)
         self.httpRequests.put((url,{
             'params': {
                 'chat_id': chat_id,
                 'text': text,
             } | kwargs
         } , callback))
+        return False, (url, "Add `urgent=True` to kwargs")
 
     @loguru.logger.catch()
     def deleteMessage(self, chat_id: int, msg_id: int):
@@ -610,6 +559,10 @@ class Autogram:
     def editMessageText(self, chat_id: int, msg_id: int, text: str, **kwargs):
         if msg_id in self.deleted_:
             return False
+        _hash = hash((chat_id, msg_id, text))
+        if _hash in self.edited_:
+            return True
+        self.edited_.add(_hash)
         #
         url = f'{self.base_url}/editMessageText'
         self.httpRequests.put((url, {
@@ -677,19 +630,21 @@ class Autogram:
             'chat_id':chat_id,
             'caption': caption,
         })
-        self.sendChatAction(chat_id,chat_actions.photo)
+        self.sendChatAction(chat_id, chat_actions.photo)
         return self.webRequest(url,params=params,files={'photo':photo_bytes})
 
     @loguru.logger.catch()
-    def sendAudio(self,chat_id: int,audio_bytes: bytes, caption: str|None = None, params: dict|None = None):
+    def sendAudio(self,chat_id: int,audio :bytes, caption: str|None = None, params: dict|None = None):
         params = params or {}
         url = f'{self.base_url}/sendAudio'
-        params.update({
+        params |= {
             'chat_id':chat_id,
             'caption': caption
-        })
-        self.sendChatAction(chat_id,chat_actions.audio)
-        return self.webRequest(url,params,files={'audio':audio_bytes})
+        }
+        if not audio:
+            raise RuntimeError('No audio bytes provided!')
+        self.sendChatAction(chat_id, chat_actions.audio)
+        return self.webRequest(url,params=params,files={'audio':audio})
 
     @loguru.logger.catch()
     def sendDocument(self,chat_id: int ,document_bytes: bytes, caption: str|None = None, params: dict|None = None):
@@ -699,7 +654,7 @@ class Autogram:
             'chat_id':chat_id,
             'caption':caption
         })
-        self.sendChatAction(chat_id,chat_actions.document)
+        self.sendChatAction(chat_id, chat_actions.document)
         return self.webRequest(url,params,files={'document':document_bytes})
 
     @loguru.logger.catch()
@@ -710,7 +665,7 @@ class Autogram:
             'chat_id':chat_id,
             'caption':caption
         })
-        self.sendChatAction(chat_id,chat_actions.video)
+        self.sendChatAction(chat_id, chat_actions.video)
         return self.webRequest(url,params,files={'video':video_bytes})
 
     @loguru.logger.catch()
@@ -740,8 +695,7 @@ class Autogram:
     @loguru.logger.catch()
     def parseFilters(self, filters: dict|None = None):
         filters = filters or {}
-        keys = list(filters.keys())
-        return json.dumps(keys)
+        return json.dumps(filters.keys())
 
     @loguru.logger.catch()
     def removeKeyboard(self, params: dict|None = None):
