@@ -2,15 +2,16 @@ from . import UpdateBase
 from typing import Dict, Callable
 
 class Notification(UpdateBase):
-    handler = None
+    name = 'notification'
 
     def __init__(self, update: Dict):
         self.notice = update
         self.autogram.logger.debug(f'Notification: {update}')
-    
+
     @classmethod
     def addHandler(cls, handler: Callable):
         cls.handler = handler
 
     def __repr__(self):
         return str(vars(self))
+
